from pathlib import Path
from .tbls import TBLSConfig
from strictyaml import load, Map, Str, Optional, Int
import json

INSERT_INTO = """


INSERT INTO {table_name} ({column_list})
VALUES
    {value_list}


"""


def formatted(data_list):
    return [f"'{x}'" if isinstance(x, str) else str(x) for x in data_list]


class Fixture:
    def __init__(self, fixture_dict, tbls_config):
        self._fix_dict = fixture_dict
        self._tbls_config = tbls_config

    def _value_list(self, table, data_dict, column_list):
        tbl = self._tbls_config._tbl_dict[table]
        sql_text = ""
        items = list(data_dict.items())

        for pk, data in items[:-1]:
            row_list = formatted(tbl.row_list(data))
            sql_text += "(" + str(pk) + "," + ",".join(row_list) + "),\n    "

        pk, data = items[-1]
        row_list = formatted(tbl.row_list(data))
        sql_text += "(" + str(pk) + "," + ",".join(row_list) + ");\n"

        return sql_text

    def sql(self):
        sql_text = ""

        for table, data_dict in self._fix_dict.items():
            column_list = self._tbls_config.column_list(table)
            sql_text += INSERT_INTO.format(
                table_name=table,
                column_list=", ".join(column_list),
                value_list=self._value_list(table, data_dict, column_list),
            )

        return sql_text


class HitchDb:
    def __init__(self, tbls_json_path: Path, fixture: Path):
        self._tbls_json_path = Path(tbls_json_path)
        self._fixture_path = Path(fixture)

        assert self._tbls_json_path.exists()
        assert self._fixture_path.exists()

    def sql(self):
        tbls_json = json.loads(self._tbls_json_path.read_text())
        assert tbls_json["driver"]["name"] == "postgres"

        tbls_config = TBLSConfig(tbls_json)

        fixture_dict = load(
            self._fixture_path.read_text(), tbls_config.strictyaml_schema()
        ).data
        fixture = Fixture(fixture_dict, tbls_config)
        return fixture.sql()



from strictyaml import Map, MapPattern, Int, Str


class PrimaryKey:
    def __init__(self, column_name, column_type):
        self.column_name = column_name
        self.column_type = column_type


class Table:
    def __init__(self, tbl_dict: dict):
        self._tbl_dict = tbl_dict

    @property
    def name(self):
        return self._tbl_dict["name"].replace("public.", "")

    def row_list(self, data):
        return [data[column["name"]] for column in self._tbl_dict["columns"][1:]]

    @property
    def columns(self) -> dict:
        return self._tbl_dict["columns"]

    def primary_key(self):
        constraints = self._tbl_dict["constraints"]
        primary_key_constraint_list = [
            constraint
            for constraint in constraints
            if constraint["type"] == "PRIMARY KEY"
        ]
        assert len(primary_key_constraint_list) == 1
        primary_key_constraint = primary_key_constraint_list[0]
        assert len(primary_key_constraint["columns"]) == 1
        pkey_name = primary_key_constraint["columns"][0]

        column_type = [
            column
            for column in self._tbl_dict["columns"]
            if column["name"] == pkey_name
        ][0]["type"]

        return PrimaryKey(
            column_name=pkey_name,
            column_type=column_type,
        )

    def _column_schema(self, column):
        if column["type"].startswith("varchar"):
            return Str()
        elif column["type"] == "integer":
            return Int()
        else:
            raise NotImplementedError()

    def column_schemas(self):
        return Map(
            {
                column["name"]: self._column_schema(column)
                for column in self._tbl_dict["columns"]
                if column["name"] != self.primary_key().column_name
            }
        )

    def strictyaml_schema(self):
        primary_key = self.primary_key()
        return MapPattern(
            Int() if primary_key.column_type == "integer" else Str(),
            self.column_schemas(),
        )


class TBLSConfig:
    def __init__(self, tbls_json: dict):
        self._tj = tbls_json

        for table in tbls_json["tables"]:
            assert table["name"].startswith("public.")

        self._tbl_dict = {
            tbl_dict["name"].replace("public.", ""): Table(tbl_dict)
            for tbl_dict in tbls_json["tables"]
        }

    def column_list(self, table_name):
        return [col["name"] for col in self._tbl_dict[table_name].columns]

    def strictyaml_schema(self):
        return Map(
            {name: tbl.strictyaml_schema() for name, tbl in self._tbl_dict.items()}
        )
